from lxml.etree import _Element, tostring
from pprint import pprint, pformat
from typing import Any, Dict, List, Optional
from datapatch import Result

from followthemoney.proxy import EntityProxy


def audit_data(data: Dict[Optional[str], Any], ignore: List[str] = []) -> None:
    """Print a row if any of the fields not ignored are still unused."""
    cleaned = {}
    for key, value in data.items():
        if key in ignore:
            continue
        if value is None or value == "":
            continue
        cleaned[key] = value
    if len(cleaned):
        pprint(cleaned)


def inspect(obj: Any) -> Optional[str]:
    """Deep-view an object for debug purposes."""
    if isinstance(obj, _Element):
        return tostring(obj, encoding="utf-8", pretty_print=True).decode("utf-8")
    if isinstance(obj, EntityProxy):
        obj = obj.to_dict()
    if isinstance(obj, Result):
        obj = repr(obj)
    return pformat(obj)



import os
from pathlib import Path
from datetime import datetime

DATA_PATH_ = os.environ.get("ZAVOD_DATA_PATH", "data")
DATA_PATH = Path(DATA_PATH_)

# Per-run timestamp
RUN_TIME = datetime.utcnow().replace(microsecond=0)

HTTP_TIMEOUT = 1200
HTTP_USER_AGENT = "Mozilla/5.0 (zavod)"
HTTP_USER_AGENT = os.environ.get("ZAVOD_HTTP_USER_AGENT", HTTP_USER_AGENT)



from typing import Optional
from normality import slugify


def join_slug(
    *parts: Optional[str],
    prefix: Optional[str] = None,
    sep: str = "-",
    strict: bool = True
) -> Optional[str]:
    sections = [slugify(p, sep=sep) for p in parts]
    if strict and None in sections:
        return None
    texts = [p for p in sections if p is not None]
    if not len(texts):
        return None
    prefix = slugify(prefix, sep=sep)
    if prefix is not None:
        texts = [prefix, *texts]
    return sep.join(texts)



from typing import Generic

from nomenklatura.entity import CE


class Sink(Generic[CE]):
    def emit(self, entity: CE) -> None:
        raise NotImplemented

    def close(self) -> None:
        pass



from typing import Union
from lxml import etree

ElementOrTree = Union[etree._Element, etree._ElementTree]


def remove_namespace(el: ElementOrTree) -> ElementOrTree:
    """Remove namespace in the passed XML/HTML document in place and
    return an updated element tree.

    If the namespaces in a document define multiple tags with the same
    local tag name, this will create ambiguity and lead to errors. Most
    XML documents, however, only actively use one namespace."""
    for elem in el.iter():
        elem.tag = etree.QName(elem).localname
        for key, value in list(elem.attrib.items()):
            local_key = etree.QName(key).localname
            if key != local_key:
                elem.attrib[local_key] = value
    etree.cleanup_namespaces(el)
    return el



from typing import Optional
from normality import collapse_spaces
from followthemoney.util import join_text

from nomenklatura.entity import CE


def make_name(
    full: Optional[str] = None,
    name1: Optional[str] = None,
    first_name: Optional[str] = None,
    given_name: Optional[str] = None,
    name2: Optional[str] = None,
    second_name: Optional[str] = None,
    middle_name: Optional[str] = None,
    name3: Optional[str] = None,
    patronymic: Optional[str] = None,
    matronymic: Optional[str] = None,
    name4: Optional[str] = None,
    name5: Optional[str] = None,
    tail_name: Optional[str] = None,
    last_name: Optional[str] = None,
) -> Optional[str]:
    """Provides a standardised way of assembling the components of a human name.
    This does a whole lot of cultural ignorance work, so YMMV."""
    full = collapse_spaces(full)
    if full is not None and len(full) > 1:
        return full
    return join_text(
        name1,
        first_name,
        given_name,
        name2,
        second_name,
        middle_name,
        name3,
        patronymic,
        matronymic,
        name4,
        name5,
        tail_name,
        last_name,
    )


def set_name_part(entity: CE, prop: str, value: Optional[str], quiet: bool) -> None:
    if value is None:
        return
    prop_ = entity.schema.get(prop)
    if prop_ is None:
        if quiet:
            return
        raise TypeError("Invalid prop: %s [value: %r]" % (prop, value))
    entity.add(prop, value, quiet=quiet)


def apply_name(
    entity: CE,
    full: Optional[str] = None,
    name1: Optional[str] = None,
    first_name: Optional[str] = None,
    given_name: Optional[str] = None,
    name2: Optional[str] = None,
    second_name: Optional[str] = None,
    middle_name: Optional[str] = None,
    name3: Optional[str] = None,
    patronymic: Optional[str] = None,
    matronymic: Optional[str] = None,
    name4: Optional[str] = None,
    name5: Optional[str] = None,
    tail_name: Optional[str] = None,
    last_name: Optional[str] = None,
    maiden_name: Optional[str] = None,
    alias: bool = False,
    name_prop: str = "name",
    is_weak: bool = False,
    quiet: bool = False,
) -> None:
    if not is_weak:
        set_name_part(entity, "firstName", given_name, quiet)
        set_name_part(entity, "firstName", first_name, quiet)
        set_name_part(entity, "secondName", second_name, quiet)
        set_name_part(entity, "fatherName", patronymic, quiet)
        set_name_part(entity, "motherName", matronymic, quiet)
        set_name_part(entity, "lastName", last_name, quiet)
        set_name_part(entity, "lastName", maiden_name, quiet)
        set_name_part(entity, "firstName", name1, quiet)
        set_name_part(entity, "secondName", name2, quiet)
        set_name_part(entity, "middleName", name3, quiet)
        set_name_part(entity, "middleName", name4, quiet)
        set_name_part(entity, "middleName", name5, quiet)
        set_name_part(entity, "lastName", tail_name, quiet)
    if alias:
        name_prop = "alias"
    if is_weak:
        name_prop = "weakAlias"
    full = make_name(
        full=full,
        name1=name1,
        first_name=first_name,
        given_name=given_name,
        name2=name2,
        second_name=second_name,
        middle_name=middle_name,
        name3=name3,
        patronymic=patronymic,
        matronymic=matronymic,
        name4=name4,
        name5=name5,
        tail_name=tail_name,
        last_name=last_name,
    )
    if full is not None and len(full):
        entity.add(name_prop, full, quiet=quiet)



from typing import Optional
from normality import slugify
from functools import cache, lru_cache
from addressformatting import AddressFormatter
from followthemoney.types import registry
from nomenklatura.entity import CE
from followthemoney.util import make_entity_id, join_text

from zavod.context import GenericZavod
from zavod.dataset import ZD


@cache
def get_formatter() -> AddressFormatter:
    return AddressFormatter()


@lru_cache(maxsize=200000)
def format_line(
    summary: Optional[str],
    po_box: Optional[str],
    street: Optional[str],
    postal_code: Optional[str],
    city: Optional[str],
    state: Optional[str],
    country_code: Optional[str],
) -> str:
    data = {
        "attention": summary,
        "house": po_box,
        "road": street,
        "postcode": postal_code,
        "city": city,
        "state": state,
    }
    return get_formatter().one_line(data, country=country_code)


def make_address(
    context: GenericZavod[CE, ZD],
    full: Optional[str] = None,
    remarks: Optional[str] = None,
    summary: Optional[str] = None,
    po_box: Optional[str] = None,
    street: Optional[str] = None,
    street2: Optional[str] = None,
    street3: Optional[str] = None,
    city: Optional[str] = None,
    place: Optional[str] = None,
    postal_code: Optional[str] = None,
    state: Optional[str] = None,
    region: Optional[str] = None,
    country: Optional[str] = None,
    country_code: Optional[str] = None,
    key: Optional[str] = None,
) -> CE:
    """Generate an address schema object adjacent to the main entity."""

    city = join_text(place, city, sep=", ")
    street = join_text(street, street2, street3, sep=", ")

    address = context.make("Address")
    address.add("full", full)
    address.add("remarks", remarks)
    address.add("summary", summary)
    address.add("postOfficeBox", po_box)
    address.add("street", street)
    address.add("city", city)
    address.add("postalCode", postal_code)
    address.add("region", region)
    address.add("state", state, quiet=True)
    address.add("country", country)
    address.add("country", country_code)

    country_code = address.first("country")
    if not address.has("full"):
        full = format_line(
            summary=summary,
            po_box=po_box,
            street=street,
            postal_code=postal_code,
            city=city,
            state=join_text(region, state, sep=", "),
            country_code=country_code,
        )
        address.add("full", full)

    full_country = registry.country.clean(full)
    if full_country is not None:
        address.add("country", full_country)
        # full = None

    # full = clean_address(full)
    address.set("full", full)

    if full:
        norm_full = slugify(full)
        hash_id = make_entity_id(country_code, norm_full, key)
        if hash_id is not None:
            address.id = f"addr-{hash_id}"
    return address



import logging
import warnings
from typing import Any, Optional
from functools import partial
from pathlib import Path
from requests import Session
from urllib3.connectionpool import InsecureRequestWarning

from zavod import settings

log = logging.getLogger(__name__)
warnings.filterwarnings("ignore", category=InsecureRequestWarning)


def make_session(user_agent: str = settings.HTTP_USER_AGENT) -> Session:
    session = Session()
    session.headers["User-Agent"] = user_agent
    session.verify = False
    session.request = partial(session.request, timeout=settings.HTTP_TIMEOUT)  # type: ignore
    return session


def fetch_file(
    session: Session,
    url: str,
    name: str,
    data_path: Path = settings.DATA_PATH,
    auth: Optional[Any] = None,
    headers: Optional[Any] = None,
) -> Path:
    """Fetch a (large) file via HTTP to the data path."""
    out_path = data_path.joinpath(name)
    if out_path.exists():
        return out_path
    log.info("Fetching: %s", url)
    out_path.parent.mkdir(parents=True, exist_ok=True)
    with session.get(url, auth=auth, headers=headers, stream=True) as res:
        res.raise_for_status()
        with open(out_path, "wb") as fh:
            for chunk in res.iter_content(chunk_size=8192):
                fh.write(chunk)
    return out_path



import sys
from threading import Lock
from typing import BinaryIO, Optional
from nomenklatura.entity import CE
from followthemoney.util import PathLike
from followthemoney.cli.util import write_entity

from zavod.sinks.common import Sink


class JSONFileSink(Sink[CE]):
    def __init__(self, path: PathLike) -> None:
        self.path = path
        self.lock = Lock()
        self.fh: Optional[BinaryIO] = None

    def emit(self, entity: CE) -> None:
        with self.lock:
            if self.fh is None:
                if str(self.path) == "-":
                    self.fh = sys.stdout.buffer
                else:
                    self.fh = open(self.path, "wb")
            write_entity(self.fh, entity)

    def close(self) -> None:
        with self.lock:
            if self.fh is not None:
                if str(self.path) != "-":
                    self.fh.close()
                self.fh = None



import json
from pathlib import Path
from typing import Any, Generic, Optional, Type, Union
from followthemoney import model
from followthemoney.schema import Schema
from nomenklatura.entity import CE
from followthemoney.util import make_entity_id, PathLike

from zavod import settings
from zavod.audit import inspect
from zavod.http import fetch_file, make_session
from zavod.sinks.common import Sink
from zavod.util import join_slug
from zavod.logs import get_logger
from zavod.dataset import ZD


class GenericZavod(Generic[CE, ZD]):
    def __init__(
        self,
        dataset: ZD,
        entity_type: Type[CE],
        sink: Optional[Sink[CE]] = None,
        data_path: Path = settings.DATA_PATH,
    ):
        self.dataset = dataset
        self.entity_type = entity_type
        self.path = data_path
        self.sink = sink
        self.log = get_logger(dataset.name)
        self.http = make_session()

    def get_resource_path(self, name: PathLike) -> Path:
        path = self.path.joinpath(name)
        path.parent.mkdir(parents=True, exist_ok=True)
        return path

    def export_metadata(self, name: PathLike = "index.json") -> Path:
        path = self.get_resource_path(name)
        with open(path, "w") as fh:
            json.dump(self.dataset.to_dict(), fh)
        return path

    def fetch_resource(
        self,
        name: str,
        url: str,
        auth: Optional[Any] = None,
        headers: Optional[Any] = None,
    ) -> Path:
        """Fetch a URL into a file located in the current run folder,
        if it does not exist."""
        return fetch_file(
            self.http,
            url,
            name,
            data_path=self.path,
            auth=auth,
            headers=headers,
        )

    def make(self, schema: Union[str, Schema]) -> CE:
        """Make a new entity with some dataset context set."""
        return self.entity_type(model, {"schema": schema})

    def make_slug(
        self, *parts: Optional[str], strict: bool = True, prefix: Optional[str] = None
    ) -> Optional[str]:
        prefix = self.dataset.prefix if prefix is None else prefix
        slug = join_slug(*parts, prefix=prefix, strict=strict)
        if slug is not None:
            return slug[:255]
        return None

    def make_id(
        self, *parts: Optional[str], prefix: Optional[str] = None
    ) -> Optional[str]:
        hashed = make_entity_id(*parts, key_prefix=self.dataset.name)
        if hashed is None:
            return None
        return self.make_slug(hashed, prefix=prefix, strict=True)

    def inspect(self, obj: Any) -> None:
        """Display an object in a form suitable for inspection."""
        text = inspect(obj)
        if text is not None:
            self.log.info(text)

    def emit(self, entity: CE) -> None:
        if self.sink is None:
            return None
        return self.sink.emit(entity)

    def close(self) -> None:
        """Flush and tear down the context."""
        self.http.close()
        if self.sink is not None:
            self.sink.close()



import click
import logging

from zavod.logs import configure_logging, get_logger

log = get_logger(__name__)


@click.group(help="Zavod data factory")
def cli() -> None:
    configure_logging(level=logging.INFO)



from setuptools import setup, find_packages

with open("README.md") as f:
    long_description = f.read()


setup(
    name="zavod",
    version="0.5.0",
    description="Data factory for followthemoney data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="data mapping identity followthemoney etl parsing",
    author="Friedrich Lindenberg",
    author_email="friedrich@opensanctions.org",
    url="https://github.com/opensanctions/zavod",
    license="MIT",
    packages=find_packages(exclude=["ez_setup", "examples", "tests"]),
    namespace_packages=[],
    include_package_data=True,
    package_data={"": ["zavod/data/*", "zavod/py.typed"]},
    zip_safe=False,
    install_requires=[
        "followthemoney >= 3.2.0, < 4.0.0",
        "nomenklatura >= 2.7.5, < 3.0.0",
        "addressformatting >= 1.3.0, < 2.0.0",
        "datapatch >= 0.2.1",
        "click >= 8.0.0, < 8.2.0",
        "requests",
        "structlog",
        "lxml",
    ],
    tests_require=[],
    entry_points={
        "console_scripts": [
            "zavod = zavod.cli:cli",
        ],
    },
    extras_require={
        "dev": [
            "wheel>=0.29.0",
            "twine",
            "mypy",
            "flake8>=2.6.0",
            "pytest",
            "pytest-cov",
            "lxml-stubs",
            "coverage>=4.1",
            "types-setuptools",
            "types-requests",
        ]
    },
)
